﻿using Android.App;
using Android.OS;
using Android.Support.V7.App;
using Android.Runtime;
using Android.Widget;
using Android.Views;
using PalmSens.Core.Simplified.Data;
using PalmSens.Core.Simplified.Android;
using System;
using System.Collections.Generic;
using SDKPlot.Android;
using System.IO;
using System.Linq;
using Android.Support.V4.App;
using Android;
using Android.Content.PM;
using Android.Support.Design.Widget;
//using SDKPlot.Android;

namespace PSSDKDataExample
{
    [Activity(Label = "@string/app_name", Theme = "@style/AppTheme", MainLauncher = true)]
    public class MainActivity : AppCompatActivity
    {
        /// <summary>
        /// The main view
        /// </summary>
        private View _view;

        /// <summary>
        /// List of permission ids used in the OnRequestPermissionResult
        /// </summary>
        private enum PermissionGroupIDs
        {
            BlueTooth = 0,
            ExternalStorage = 1
        }

        /// <summary>
        /// The first operation text
        /// </summary>
        TextView _txtOperation1;

        /// <summary>
        /// The second operation text
        /// </summary>
        TextView _txtOperation2;

        /// <summary>
        /// The third operation text
        /// </summary>
        TextView _txtOperation3;

        /// <summary>
        /// The perform operation button
        /// </summary>
        Button _btnPerformOperation;

        /// <summary>
        /// The SDKPlot android plot object
        /// </summary>
        Plot _plot;

        /// <summary>
        /// The next data operation to perform (always start with loading the data)
        /// </summary>
        private Steps _nextStep = Steps.LoadData;

        private enum Steps
        {
            LoadData = 1,
            Smooth = 2,
            MovingAverageBaseline = 3,
            SubtractBaseline = 4,
            DetectPeaks = 5,
            NewChargeCurve = 6,
            DifferntiateCurve = 7,
            DisplayBodePlot = 8
        }

        private string[] operations = new string[]
        {
            "1. Load PSSession File...",
            "2. Smooth Curve...",
            "3. Determine Moving Average Baseline...",
            "4. Subtract Baseline...",
            "5. Detect Peaks...",
            "6. New Charge over Time Curve...",
            "7. Differentiate Curve...",
            "8. Display Bode Plot of EIS Measurement..."
        };

        /// <summary>
        /// List that will hold references to the measurements in the Example PSSession file
        /// </summary>
        private List<SimpleMeasurement> _measurements;

        /// <summary>
        /// The active SimpleMeasurement
        /// </summary>
        private SimpleMeasurement _activeMeasurement;

        /// <summary>
        /// The active SimpleCurve
        /// </summary>
        private SimpleCurve _activeCurve;

        /// <summary>
        /// The baseline SimpleCurve
        /// </summary>
        private SimpleCurve _baselineCurve;

        protected override void OnCreate(Bundle bundle)
        {
            base.OnCreate(bundle);

            // Set our view from the "main" layout resource
            SetContentView(Resource.Layout.activity_main);

            _view = FindViewById<LinearLayout>(Resource.Id.mainLayout);

            //Get references to textview controls
            _txtOperation1 = FindViewById<TextView>(Resource.Id.txtOperation1);
            _txtOperation1.Text = operations[0];
            _txtOperation1.SetTextColor(Android.Graphics.Color.Black);
            _txtOperation2 = FindViewById<TextView>(Resource.Id.txtOperation2);
            _txtOperation2.Text = operations[1];
            _txtOperation2.SetTextColor(Android.Graphics.Color.LightGray);
            _txtOperation3 = FindViewById<TextView>(Resource.Id.txtOperation3);
            _txtOperation3.Text = operations[2];
            _txtOperation3.SetTextColor(Android.Graphics.Color.LightGray);

            //Get reference to button control
            _btnPerformOperation = FindViewById<Button>(Resource.Id.btnPerformOperation);
            _btnPerformOperation.Click += _btnPerformOperation_Click;

            ////Get reference to SDKPlot plot control
            _plot = FindViewById<Plot>(Resource.Id.plot);

            InitPlot(); //Resets and initiates the plot control
        }

        /// <summary>
        /// Called when app process is resumed and after OnStart when app is started
        /// </summary>
        protected override void OnResume()
        {
            base.OnResume();

            RequestDangerousPermissions(); //Necessary as of Android version 23
        }

        /// <summary>
        /// Request permissions that require user confirmation since android 23
        /// </summary>
        private void RequestDangerousPermissions()
        {
            //Only request permissions on Android versions after 22
            if (Build.VERSION.SdkInt >= BuildVersionCodes.M)
            {
                //Check if external storage permissions are granted, technically only required when accessing files from external storage (this example accesses a file stored in its Assets which does not require this permission)
                if (ActivityCompat.CheckSelfPermission(this, Manifest.Permission.ReadExternalStorage) != (int)Permission.Granted || ActivityCompat.CheckSelfPermission(this, Manifest.Permission.WriteExternalStorage) != (int)Permission.Granted)
                {
                    //Request external storage permissions
                    ActivityCompat.RequestPermissions(this, new string[] { Manifest.Permission.ReadExternalStorage, Manifest.Permission.WriteExternalStorage }, (int)PermissionGroupIDs.ExternalStorage);
                }
            }
        }

        /// <summary>
        /// Callback received when user permission request has been completed
        /// </summary>
        /// <param name="requestCode"></param>
        /// <param name="permissions"></param>
        /// <param name="grantResults"></param>
        public override void OnRequestPermissionsResult(int requestCode, string[] permissions, [GeneratedEnum] Permission[] grantResults)
        {
            switch ((PermissionGroupIDs)requestCode)
            {
                case PermissionGroupIDs.ExternalStorage:
                    if (grantResults.Length == 2 && grantResults[0] == Permission.Granted && grantResults[1] == Permission.Granted)
                    {
                        //Permission granted
                    }
                    else
                    {
                        //Permission denied
                        if (ActivityCompat.ShouldShowRequestPermissionRationale(this, Manifest.Permission.ReadExternalStorage) || ActivityCompat.ShouldShowRequestPermissionRationale(this, Manifest.Permission.WriteExternalStorage))
                        {
                            //Inform user that BlueTooth 
                            Snackbar.Make(_view, "Without granting permission to external storage load and saving from and to the SD card will not be possible.", Snackbar.LengthLong).Show();
                        }
                        else
                        {
                            //Occurs when user previously denied permission and selected never ask again
                            Snackbar.Make(_view, "Without granting permission to external storage load and saving from and to the SD card will not be possible.\nTo use the external storage you need to manualy grant the external read and write for this app in the android settings", Snackbar.LengthLong).Show();
                        }
                    }
                    break;
                case PermissionGroupIDs.BlueTooth: //Not used in this example
                default:
                    base.OnRequestPermissionsResult(requestCode, permissions, grantResults);
                    break;
            }
        }

        /// <summary>
        /// Initializes the plot control.
        /// </summary>
        private void InitPlot()
        {
            _plot.ClearAll(); //Clear all curves and data from plot
            //Set the Axis labels
            _plot.XAxisLabel = "V";
            _plot.YAxisLabel = "µA";
            _plot.AxisTextColor = OxyPlot.OxyColors.White;
            _plot.AxesColor = OxyPlot.OxyColors.LightGray;
            _plot.AddData("", new double[0], new double[0]); //Add a empty data array to draw an empty plot
        }

        /// <summary>
        /// Displays a Toast message.
        /// </summary>
        /// <param name="message">The message.</param>
        private void DisplayMessage(string message)
        {
            Toast toast = Toast.MakeText(this, message, ToastLength.Short);
            toast.Show();
        }

        /// <summary>
        /// Handles the Click event of the _btnPerformOperation control.
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="System.EventArgs"/> instance containing the event data.</param>
        /// <exception cref="System.NotImplementedException"></exception>
        private void _btnPerformOperation_Click(object sender, System.EventArgs e)
        {
            PerformOperation();
        }

        /// <summary>
        /// Performs the pending operation.
        /// </summary>
        private void PerformOperation()
        {
            if (_nextStep == Steps.LoadData)
            {
                //Load measurements from the Example PSSession File
                //Streams are used when loading files stored in your project's Assets folder
                using (StreamReader file = new StreamReader(Assets.Open("Example.pssession")))
                    _measurements = SimpleLoadSaveFunctions.LoadMeasurements(file);
                //Alternatively you can use a filepath to load a file instead of a Stream.
                //Common file paths can be obtained from the Environment class, i.e. Environment.ExternalStorageDirectory.
                if (_measurements.Count == 4)
                {
                    DisplayMessage("Example.pssession succesfully loaded.");
                }
                else
                {
                    DisplayMessage("Example file empty...");
                    DisplayMessage("Please make sure the Example.pssession file is present in the project folder.");
                    return;
                }

                //Get the measurement named DPV Example
                //Also possible without LINQ: SimpleMeasurement dpvExample = _measurements[2]
                SimpleMeasurement dpvExample = _measurements.Where(meas => meas.Title == "DPV Example").FirstOrDefault();
                _activeMeasurement = dpvExample;

                //Plot the first and only curve in the dpvExample measurement
                _activeCurve = dpvExample.SimpleCurveCollection[0];
                _plot.AddSimpleCurve(_activeCurve);

                _txtOperation1.Text += " Done";
                _txtOperation1.SetTextColor(Android.Graphics.Color.LightGray);
                _txtOperation2.SetTextColor(Android.Graphics.Color.White);
                IncrementStep(); //Increment the _nextStep property
            }
            else if (_nextStep == Steps.Smooth)
            {
                //Smooth the active SimpleCurve defined in the previous step
                //Smoothes the data on the Y-axis with the Savitzky-Golay filter.
                SimpleCurve smoothedCurve = _activeCurve.Smooth(SmoothLevel.High);

                DisplayMessage("Curve Smoothed.");

                //Add the smoothed curve to the plot
                _plot.AddSimpleCurve(smoothedCurve);

                _txtOperation1.Text = $"{operations[1]} DONE";
                _txtOperation2.Text = operations[2];
                _txtOperation3.Text = operations[3];
                IncrementStep(); //Increment the _nextStep property
            }
            else if (_nextStep == Steps.MovingAverageBaseline)
            {
                _plot.ClearAll(); //Clear the plot

                //Get the measurement named SWV Example
                //Also possible without LINQ: SimpleMeasurement dpvExample = _measurements[3]
                SimpleMeasurement swvExample = _measurements.Where(meas => meas.Title == "SWV Example").FirstOrDefault();
                _activeMeasurement = swvExample;

                //Get the first and only curve in the SWV Example measurement
                _activeCurve = swvExample.SimpleCurveCollection[0];
                _plot.AddSimpleCurve(_activeCurve);

                //Determine and plot the moving average baseline curve
                _baselineCurve = _activeCurve.MovingAverageBaseline();
                _plot.AddSimpleCurve(_baselineCurve);

                DisplayMessage("SWV Example Curve Moving Average Baseline Determined.");

                _txtOperation1.Text = $"{operations[2]} DONE";
                _txtOperation2.Text = operations[3];
                _txtOperation3.Text = operations[4];
                IncrementStep(); //Increment the _nextStep property
            }
            else if (_nextStep == Steps.SubtractBaseline)
            {
                //Subtract the baseline from the measurement and plot the result
                _activeCurve = _activeCurve.Subtract(_baselineCurve); //Replace the activeCurve reference with the result
                _plot.AddSimpleCurve(_activeCurve);

                DisplayMessage("Baseline Curve Subtracted From SWV Example Curve.");

                _txtOperation1.Text = $"{operations[3]} DONE";
                _txtOperation2.Text = operations[4];
                _txtOperation3.Text = operations[5];
                IncrementStep(); //Increment the _nextStep property
            }
            else if (_nextStep == Steps.DetectPeaks)
            {
                _plot.ClearAll(); //Clear the plot

                //Add the last step's result to the plot and detect its peaks
                _plot.AddSimpleCurve(_activeCurve);
                _activeCurve.DetectPeaks();

                DisplayMessage("Peaks Detected in the Baseline Corrected Curve.");

                _txtOperation1.Text = $"{operations[4]} DONE";
                _txtOperation2.Text = operations[5];
                _txtOperation3.Text = operations[6];
                IncrementStep(); //Increment the _nextStep property
            }
            else if (_nextStep == Steps.NewChargeCurve)
            {
                _plot.ClearAll(); //Clear the plot

                //Get the measurement named CV Example
                //Also possible without LINQ: SimpleMeasurement dpvExample = _measurements[0]
                SimpleMeasurement cvExample = _measurements.Where(meas => meas.Title == "CV Example").FirstOrDefault();
                _activeMeasurement = cvExample;

                //Get Charge over Time curves from the measurement and plot them
                List<SimpleCurve> chargeCurves = cvExample.NewSimpleCurve(PalmSens.Data.DataArrayType.Time, PalmSens.Data.DataArrayType.Charge, "Charge/Potential");
                _plot.AddSimpleCurves(chargeCurves);
                _activeCurve = chargeCurves[0];

                DisplayMessage("Obtained Charge over Time curve for each scan in the CV Example.");

                _txtOperation1.Text = $"{operations[5]} DONE";
                _txtOperation2.Text = operations[6];
                _txtOperation3.Text = operations[7];
                IncrementStep(); //Increment the _nextStep property
            }
            else if (_nextStep == Steps.DifferntiateCurve)
            {
                _plot.ClearAll(); //Clear the plot

                //Plot the first charge over time scan from the previous step
                _plot.AddSimpleCurve(_activeCurve);

                //Differentiate and plot the result on the secondary y-axis
                SimpleCurve diffCharge = _activeCurve.Differentiate();
                _plot.AddSimpleCurve(diffCharge, true); //Plot on the secondary y-axis
                _plot.YAxisSecondaryLabel = "dC(µC)/dt(s)"; //Set the unit for the secondary Y-Axis

                DisplayMessage("Differentiated the curve");

                _txtOperation2.Text = $"{operations[6]} DONE";
                _txtOperation2.SetTextColor(Android.Graphics.Color.LightGray);
                _txtOperation3.SetTextColor(Android.Graphics.Color.White);
                IncrementStep(); //Increment the _nextStep property
            }
            else if (_nextStep == Steps.DisplayBodePlot)
            {
                _plot.ClearAll(); //Clear the plot

                //Get the measurement named EIS Example
                //Also possible without LINQ: SimpleMeasurement dpvExample = _measurements[3]
                SimpleMeasurement eisExample = _measurements.Where(meas => meas.Title == "EIS Example").FirstOrDefault();
                _activeMeasurement = eisExample;

                //Get the curves for the bode plot from the measurement
                SimpleCurve impedance = (eisExample.NewSimpleCurve(PalmSens.Data.DataArrayType.Frequency, PalmSens.Data.DataArrayType.Z, "Impedance"))[0];
                SimpleCurve phase = (eisExample.NewSimpleCurve(PalmSens.Data.DataArrayType.Frequency, PalmSens.Data.DataArrayType.Phase, "-Phase"))[0];

                //Plot the curves on the primary and secondary y-axes and set the x-axis to logarithmic display mode
                _plot.AddSimpleCurve(impedance);
                _plot.AddSimpleCurve(phase, true); //Plot on the secondary y-axis
                //Set the axis to logarithmic display mode
                _plot.XAxisType = SDKPlot.AxisType.Logarithmic;
                _plot.YAxisType = SDKPlot.AxisType.Logarithmic;

                _txtOperation3.Text = $"{operations[7]} DONE";
                _txtOperation3.SetTextColor(Android.Graphics.Color.LightGray);
                _btnPerformOperation.Enabled = false;
            }
        }

        /// <summary>
        /// Increments the next step property.
        /// </summary>
        private void IncrementStep()
        {
            int nextStepIndex = (int)_nextStep + 1;
            _nextStep = (Steps)nextStepIndex;
        }
    }
}